import Mathlib.Algebra.Algebra.Basic

import Mathlib.Tactic.LibrarySearch

/-! # The typeclass `CommAlgebra` -/

/-- A commutative algebra is an algebra (see `Algebra`) in which multiplication is
commutative. -/
class CommAlgebra (R : Type u) [CommSemiring R] (A : Type v) [Semiring A]
    extends Algebra R A where
  /-- Multiplicative is commutative in a commutative algebra. -/
  mul_comm (a b : A) : a * b = b * a

namespace CommAlgebra

variable (R : Type u)
section RCommSemiring variable [CommSemiring R] -- declare R as a `CommSemiring`

variable (A : Type v)
section ASemiring variable [Semiring A] -- declare A as a `Semiring`

instance : Coe (CommAlgebra R A) (Algebra R A) where
  coe := (·.toAlgebra)

variable {R A} in
/-- Two commutative R-algebra structures on A are equal if the algebra map from R
to A is the same for both. -/
@[ext] protected theorem algebraMap_ext : {inst₁ inst₂ : CommAlgebra R A} →
    (∀ r, (haveI := inst₁; algebraMap R A r) = (haveI := inst₂; algebraMap R A r)) →
    inst₁ = inst₂
| .mk (toAlgebra := alg₁) comm₁, .mk (toAlgebra := alg₂) comm₂, h => 
  congrArg₂ (CommAlgebra.mk (toAlgebra := ·) ·)
    (Algebra.algebra_ext alg₁ alg₂ h)
    (proofIrrel comm₁ comm₂)

-- TODO ext for smul?

/- Marking this as an instance leads to a "synthesization order" error.

I think this is because the output type does not specify R, so this would lead
to a search for `CommAlgebra` instances for _every_ `CommSemiring R`.
There's something similar for `Algebra`. -/

variable {R A} in
/-- Any commutative algebra is a commutative semiring.

This is not marked as an instance; it must be manually used. -/
def toCommSemiring (self : CommAlgebra R A) : CommSemiring A := { self with }

end ASemiring section ACommSemiring variable [CommSemiring A] -- redeclare A as a `CommSemiring`

section variable {R A} variable (f : R →+* A)

/-- Creating a commutative algebra from a morphism to a commutative semiring. 

The function `RingHom.toAlgebra` creates an algebra, but in fact the algebra is
always a commutative algebra if `A` is a commutative semiring. -/
def _root_.RingHom.toCommAlgebra : CommAlgebra R A where
  __ := f.toAlgebra
  mul_comm := CommSemiring.mul_comm

lemma _root_.RingHom.toCommAlgebra_toCommSemiring :
    f.toCommAlgebra.toCommSemiring = ‹CommSemiring A› :=
  rfl

lemma _root_.RingHom.toCommAlgebra_toRingHom :
    f.toCommAlgebra.toRingHom = f :=
  rfl

end

/-- A commutative R-algebra structure on A is equivalent to a ring hom `R →+* A`. -/
def equiv_RingHom : CommAlgebra R A ≃ (R →+* A) where
  toFun := (·.toRingHom)
  invFun := RingHom.toCommAlgebra
  left_inv _ := by ext; rfl
  right_inv := RingHom.toCommAlgebra_toRingHom

end ACommSemiring               -- end scope of A's typeclasses

end RCommSemiring section RCommRing variable [CommRing R] -- redeclare R as `CommRing`

variable {R A} [Semiring A] in
/-- Any commutative algebra over a commutative ring is a commutative ring.

This is not marked as an instance; it must be manually used. -/
def toCommRing (self : CommAlgebra R A) : CommRing A where
  __ := self
  __ := Algebra.semiringToRing R

/- Proved but waiting until the PR https://github.com/leanprover-community/mathlib4/pull/9511 is
sorted out and merged. -/
@[ext] theorem CommRing.ext {R} ⦃inst₁ inst₂ : CommRing R⦄
    (h_add : ∀ x y : R, (letI := inst₁; x + y : R) = (letI := inst₂; x + y))
    (h_mul : ∀ x y : R, (letI := inst₁; x * y : R) = (letI := inst₂; x * y)) :
    inst₁ = inst₂ :=
  sorry

lemma _root_.RingHom.toCommAlgebra_toCommRing {A : Type v} [CommRing A]
    (f : R →+* A) : f.toCommAlgebra.toCommRing = ‹CommRing A› :=
  by ext <;> dsimp

end RCommRing                   -- end scope of R's typeclasses

end CommAlgebra
